"""
TestParser class definition
"""

import unittest
from sodapclient.Parser import Parser


class TestParser(unittest.TestCase):

    """
    Test class for Parser class.
    """

    def setUp(self):

        # Create the test string
        self.start_str = 'StartsHere'
        self.data_str = """

blah...
{

    indent stuff...

""" + self.start_str + """{
    DataType1 x[x = 123];
    DataType2 y[y = 456];

    Grid {
     ARRAY:
       DataType3 z[x = 123][y = 456];
     MAPS:
       DataType1 x[x = 123];

       DataType2 y[y = 456];
    } z;

} TestString;

blah...



"""

        self.data_lines = ['blah...',
                           '{', '    indent stuff...',
                           self.start_str + '{',
                           '    DataType1 x[x = 123];',
                           '    DataType2 y[y = 456];',
                           '    Grid {',
                           '     ARRAY:',
                           '       DataType3 z[x = 123][y = 456];',
                           '     MAPS:', '       DataType1 x[x = 123];',
                           '       DataType2 y[y = 456];',
                           '    } z;',
                           '} TestString;', 'blah...']

    def tearDown(self):
        pass

    def test_constructor(self):

        """
        Test the Parser constructor.
        """

        parser = Parser()
        self.assertTrue(type(parser) == Parser)
        return parser

    def test_find_indent_level(self):

        """
        Check that the correct indent level is found.
        """

        parser = self.test_constructor()
        indt = parser.find_indent_level('     abc...  ')
        self.assertEqual(indt, 5)

    def test_find_start(self):

        """
        Check that the correct start line is found.
        """

        parser = self.test_constructor()
        parser.find_start(self.data_str, self.start_str)
        self.assertEqual(parser.data_lines, self.data_lines)
        self.assertEqual(parser.lnum, 3)
        self.assertEqual(parser.indts, [0])
        return parser

    def test_check_line(self):

        """
        Check that lines are handled correctly.
        """

        parser = self.test_find_start()

        do_line = parser.check_line()
        self.assertEqual(do_line, True)
        self.assertEqual(parser.indts, [0, 4])
        self.assertEqual(parser.finished, False)

        do_line = parser.check_line()
        self.assertEqual(do_line, True)
        self.assertEqual(parser.indts, [0, 4])
        self.assertEqual(parser.finished, False)

        parser.lnum = len(parser.data_lines) - 3
        do_line = parser.check_line()
        self.assertEqual(do_line, False)
        self.assertEqual(parser.indts, [0])
        self.assertEqual(parser.finished, True)

if __name__ == "__main__":
    unittest.main()



from .Handler import Handler
from .ProxyDict import ProxyDict
from .Definitions import Definitions
from .Parser import Parser
from .DDSParser import DDSParser
from .DASParser import DASParser
from .VariableLoader import VariableLoader



"""
DDSParser class definition
"""

from .Parser import Parser
from .Definitions import Definitions


class DDSParser(Parser):

    """
    Dataset Descriptor Response (dds) parser class.

    Given a DDS string read from a URL, it parses the string to set up
    a DDS dictionary. The dictionary is of form:
    {name : [type, [dims], [assoc_names]]}
    where name is the variable name,
    dims is a list of dimensions,
    assoc_names is a list of associated names.

    See Handler class for comments on constructor handling.

    A single data set is included, i.e. if the URL describes more than one,
    only the first will be included.
    """

    def __init__(self):

        super().__init__()
        self.dtype = 'Dataset Descriptor Structure (DDS)'

        self.atomic_defs = list(Definitions.atomics.keys())
        self.constructor_defs = Definitions.constructors

        self.dataset_name = 'Undefined'

    def parse(self, dds_str):

        """
        Parse the DDS.
        args...
            dds_str: DDS string
        """

        self.find_start(dds_str, Definitions.dataset)

        # Main loop
        while (not self.finished) and \
              (self.lnum < len(self.data_lines) - 1) and (len(self.indts) > 0):
            do_line = self.check_line()
            if do_line:
                self.process_line()

        if self.finished:
            # Exclude trailing semicolon...
            self.dataset_name = self.data_lines[self.lnum].split()[-1][:-1]

    def process_line(self):

        """
        Process a line.
        """

        var_type = None
        next_line = self.data_lines[self.lnum]
        segs = next_line.split()
        for adef in self.atomic_defs:
            if segs[0] == adef:
                var_type = adef
                self.read_atomic(next_line, var_type)
                break
        if var_type is None:  # Not an atomic type
            for adef in self.constructor_defs:
                if segs[0] == adef:
                    var_type = adef
                    self.read_constructor(next_line, var_type)
                    break

    def read_atomic(self, line, var_type):

        """
        Read an atomic variable.
        args...
            line: line
            var_type: variable type
        """

        # Will still work even if dimensionless...
        var_name = line.split()[1].split('[')[0]
        segs = line.split('=')
        num_dims = len(segs) - 1
        # Remove trailing semicolon...
        if num_dims == 0:
            var_name = var_name[:-1]
        dims = [None] * num_dims
        assoc_names = [None] * num_dims
        for idim in range(num_dims):
            # Remove trailing space...
            assoc_names[idim] = segs[idim].split('[')[-1][:-1]
            dims[idim] = int(segs[idim + 1].split(']')[0])
        self.data[var_name] = [var_type, dims, assoc_names]

    def read_constructor(self, line, var_type):

        """
        Reading constructor variables not implemented
        (all variables handled as atomics).
        """

        pass

    def print_dds(self, dataset_name=None, dds=None):

        """
        Print the DDS to the console. Dataset name and DDS
        can be passed in externally for convenience.
        kwargs...
            dataset_name: dataset name (string)
            dds: DDS (dictionary)
        """

        if dataset_name is None:
            dataset_name = self.dataset_name
        if dds is None:
            dds = self.data
        self.print_data(self.dtype, dds)
        print('Dataset name :', dataset_name, '\n')
        for var in dds.keys():
            print('Variable name :', var)
            print('dtype :', dds[var][0])
            print('Dimensions :', dds[var][1])
            print('Associated names :', dds[var][2], '\n')

    def print_dds_to_file(self, file_name, dataset_name=None, dds=None):

        """
        Print the DDS to a file. Dataset name and DDS
        can be passed in externally for convenience.
        args...
            file_name: file name (string)
        kwargs...
            dataset_name: dataset name (string)
            dds: DDS (dictionary)
        """

        if dataset_name is None:
            dataset_name = self.dataset_name
        if dds is None:
            dds = self.data
        self.print_data_to_file(self.dtype, dds, file_name)
        file_name.write('Dataset name : ' + dataset_name + '\n\n')
        for var in dds.keys():
            file_name.write('Variable name : ' + var + '\n')
            file_name.write('dtype : ' + dds[var][0] + '\n')
            file_name.write('Dimensions : ' + str(dds[var][1]) + '\n')
            file_name.write('Associated names : ' + str(dds[var][2]) + '\n\n')



"""
TestDASParser class definition
"""

import os
import unittest
from ..DASParser import DASParser


class TestDASParser(unittest.TestCase):

    """
    Test class for DASParser class.
    """

    def setUp(self):

        # Create the test das string
        self.das_str = """
Attributes {
    lat {
        String name "Latitude";
        String units "deg";
        Float32 minval -90;
        Float32 maxval 90;
    }
    long {
        String name "Longitude";
        String units "deg";
        Float32 minval 0;
        Float32 maxval 360;
    }
    depth {
        String name "Water Depth";
        String units "m";
        Int32 posup -1;
    }
}
"""

        self.das = {'lat': ['String name "Latitude"', 'String units "deg"',
                            'Float32 minval -90', 'Float32 maxval 90'],
                    'long': ['String name "Longitude"', 'String units "deg"',
                             'Float32 minval 0', 'Float32 maxval 360'],
                    'depth': ['String name "Water Depth"', 'String units "m"',
                              'Int32 posup -1']
                    }

        self.test_file_name = 'dasparser_out.txt'

    def tearDown(self):
        if os.path.exists(self.test_file_name):
            os.remove(self.test_file_name)

    def test_constructor(self):

        """
        Test the DASParser constructor.
        """

        parser = DASParser()
        self.assertTrue(type(parser) == DASParser)

    def test_parse(self):

        """
        Test the parse method.
        """

        das_parser = DASParser()
        das_parser.parse(self.das_str)
        self.assertEqual(self.das, das_parser.data)

    def test_print(self):

        """
        Test the print method.
        """

        das_parser = DASParser()
        das_parser.parse(self.das_str)
        das_parser.print_das()

    def test_print_to_file(self):

        """
        Test the print to file method.
        """

        das_parser = DASParser()
        das_parser.parse(self.das_str)
        test_file = open(self.test_file_name, 'wt')
        das_parser.print_das_to_file(file_name=test_file)
        self.assertEqual(os.path.exists(self.test_file_name), True)

if __name__ == "__main__":
    unittest.main()



"""
DASParser class definition
"""

from .Parser import Parser
from .Definitions import Definitions


class DASParser(Parser):

    """
    DASParser class. Extracts and stores the attributes within a DAS string.
    """

    def __init__(self):

        super().__init__()
        self.dtype = 'Dataset Attribute Structure (DAS)'

    def parse(self, das_str):

        """
        Parse the DAS.
        args...
            das_str: DAS string.
        """

        self.find_start(das_str, Definitions.attributes)

        attr_ind = -1  # Attribute index counter

        # Main loop
        while (not self.finished) & (self.lnum < len(self.data_lines) - 1) & \
              (len(self.indts) > 0):
            do_line = self.check_line()
            if do_line:  # New Attribute
                attr_ind += 1
                attr_name = self.data_lines[self.lnum].split()[0]
                attr_data = []
                while self.check_line():
                    # Remove trailing semicolon
                    attr_data.append(self.data_lines[self.lnum].lstrip()[:-1])
                self.data[attr_name] = attr_data

    def print_das(self, das=None):

        """
        Print the DAS to the console. DAS can be passed in externally for convenience.
        kwargs...
            das: DAS (dictionary)
        """

        if das is None:
            das = self.data
        self.print_data(self.dtype, das)
        for var in das.keys():
            print('Variable :', var)
            print('Attributes...')
            for l in das[var]:
                print(l)
            print()

    def print_das_to_file(self, file_name, das=None):

        """
        Print the DAS to a file. DAS can be passed in externally for convenience.
        args...
            file_name: file name (string)
        kwargs...
            das: DAS (dictionary)
        """

        if das is None:
            das = self.data
        self.print_data_to_file(self.dtype, das, file_name)
        for var in das.keys():
            file_name.write('Variable : ' + var + '\n')
            file_name.write('Attributes...\n')
            for l in das[var]:
                file_name.write(l + '\n')
            file_name.write('\n')



from setuptools import setup, find_packages

setup(name='sodapclient',
      version='0.1.1',
      description='A simple OpenDAP client library',
      author='Systems Engineering & Assessment Ltd.',
      author_email='Marcus.Donnelly@sea.co.uk',
      url='https://bitbucket.org/sea_dev/sodapclient',
      license='MIT',
      classifiers=['Development Status :: 3 - Alpha',
                   'Intended Audience :: Science/Research',
                   'Intended Audience :: Developers',
                   'License :: OSI Approved :: MIT License',
                   'Programming Language :: Python :: 3',
                   'Topic :: Scientific/Engineering'
                   ],
      keywords=['OpenDAP',
                'Environment',
                'Science'
                ],
      packages=find_packages(),
      install_requires=['numpy >= 1.12'],
      package_data={'sodapclient.Examples': ['proxyserver.txt'],
                    },
      )



"""
Example use of sodapclient.
To run this example at the command line enter: python3 Example.py
"""

import numpy

from sodapclient import Handler

# Define the URL (one of the datasets available at the official
# OpenDAP test server):
url = 'http://test.opendap.org/opendap/hyrax/data/nc/sst.mnmean.nc.gz.html'

# Initialise handler, which will also get the DDS & DAS
handler = Handler(url, proxy_file_name='proxyserver.txt')

# We'll access the SST data over the last 2 times, the last 3 lats and the
# last 4 longs available...

# Set up the arrays for the required dimensions, shape 1x3 for map variables
# and nx3 for data variables with n dimensions
time_dims = numpy.array([[1800, 10, 1810]])  # Time indices 1800 and 1810
lat_dims = numpy.array([[70, 2, 74]])  # Latitude indices 70, 72 and 74
lon_dims = numpy.array([[80, 1, 83]])  # Longitude indices 80 to 83
sst_dims = numpy.array([time_dims[0], lat_dims[0], lon_dims[0]])

byte_order = '>'  # Big Endian

# Get the data and the dimension variables

handler.get_variable('time', time_dims, byte_order)
handler.get_variable('lat', lat_dims, byte_order)
handler.get_variable('lon', lon_dims, byte_order)
handler.get_variable('sst', sst_dims, byte_order)

handler.print()  # Print the DDS, DAS and data to screen

# Then get access to any variable from the handler's variables dictionary, e.g.

sst = handler.variables['sst']
print('\nSST...\n', sst)



"""
Variable definitions
"""

import numpy


class Definitions:

    """
    Simple 'static' class to provide OpenDAP variable definitions.
    Includes mapping to NumPy variable types where available.
    """

    atomics = {'Byte': numpy.dtype(numpy.byte),
               'Int16': numpy.dtype(numpy.int32),  # Int16->int32 in practice?
               'UInt16': numpy.dtype(numpy.uint32),  # Same assumed for UInt16
               'Int32': numpy.dtype(numpy.int32),
               'UInt32': numpy.dtype(numpy.uint32),
               'Float32': numpy.dtype(numpy.float32),
               'Float64': numpy.dtype(numpy.float64),
               'String': numpy.dtype(numpy.byte),  # Retrieve as byte sequence
               'URL': numpy.dtype(numpy.byte)  # Same as String
               }

    constructors = ['Dataset', 'Array', 'Structure',
                    'Grid', 'Sequence', 'Maps']
    # NB can be in all upper case as well

    dataset = 'Dataset'

    attributes = 'Attributes'
